
#ifndef CSR_TRI_SOLVE_GPU_H
#define CSR_TRI_SOLVE_GPU_H

// -----------------------------------------------------------------------------

#include "uplo_type.h"
#include "csr_matrix_data.h"
#include "csr_tri_solve_part_cpu.h"
#include "csr_spmv_part_vector_gpu.h"
#include "cuda_memory.h"
#include <vector>

// -----------------------------------------------------------------------------

const unsigned int CSR_TRI_SOLVE_GPU_NUM_VORDER = 2;

// -----------------------------------------------------------------------------


// solves using a given block size.  diagonal blocks are inverted on cpu.
// A, b, x reside in CPU memory.  A_gpu resides in CPU memory.
template<typename index_type, typename value_type>
bool csr_tri_solve_gpu_hybrid(const csr_matrix_data<index_type, value_type>& A,
			      index_type* nnzPerRowDiagBlock,
                              uplo_enum uplo_type,
                              value_type* b,
                              value_type* x,
                              const csr_matrix_data<index_type, value_type>& A_gpu,
                              cuda_memory<value_type>& b_gpu,
                              cuda_memory<value_type>& x_gpu,
                              index_type block_size,
                              index_type num_vecs,
                              unsigned int vorder);

template<typename index_type, typename value_type>
int csr_tri_solve_gpu_hybrid(const csr_matrix_data<index_type, value_type>* A,
			     index_type* nnzPerRowDiagBlock,
			     uplo_enum uplo_type,
			     value_type* b,
			     value_type* x,
			     const csr_matrix_data<index_type, value_type>* A_gpu,
			     value_type* b_gpu,
			     value_type* x_gpu,
			     index_type block_size,
			     index_type num_vecs,
			     unsigned int vorder);

// -----------------------------------------------------------------------------
// -----------------------------------------------------------------------------

template<typename index_type, typename value_type>
bool csr_tri_solve_gpu_hybrid(const csr_matrix_data<index_type, value_type>& A,
			      index_type* nnzPerRowDiagBlock,
                              uplo_enum uplo_type,
                              value_type* b,
                              value_type* x,
                              const csr_matrix_data<index_type, value_type>& A_gpu,
                              cuda_memory<value_type>& b_gpu,
                              cuda_memory<value_type>& x_gpu,
                              index_type block_size,
                              index_type num_vecs,
                              unsigned int vorder)
{
  bool success = false;
  
  const index_type num_rows = A.num_rows;
  
  const index_type next_vec_offset = num_rows;
  
  if( num_rows == A.num_cols &&
      num_rows > 0 &&
      num_rows == A_gpu.num_rows &&
      num_rows == A_gpu.num_cols &&
      A.num_entries == A_gpu.num_entries &&
      (uplo_type == uplo_up_only || uplo_type == uplo_lo_only) &&
      x &&
      b &&
      block_size > 0 &&
      num_vecs > 0 &&
      vorder < CSR_TRI_SOLVE_GPU_NUM_VORDER)
    {
      block_size = std::min(block_size, num_rows);

      const index_type num_blocks = (num_rows + block_size - 1)/block_size;

      // using this single bool to have a single code for both up and lo cases.
      const bool up = (uplo_type == uplo_up_only);
      
      index_type i_b = (up ? num_blocks : 0);

      while((up && i_b > 0) || (!up && i_b < num_blocks))
	{
	  if(up)
	    --i_b;
	  
	  const index_type row_start = i_b * block_size;
	  const index_type row_end   = std::min(row_start + block_size, num_rows);
	  
	  if((!up && (i_b > 0)) || (up && (i_b < num_blocks - 1)))
	    {
	      const index_type part_col_start   = up ? row_end  : 0;
	      const index_type part_col_end     = up ? num_rows : row_start;
	      const index_type part_num_entries = A.row_offsets[row_end] - A.row_offsets[row_start];

	      const unsigned int spmv_vorder = 0; // Seems like 1 does not make it faster...
	      success = csr_spmv_sub_part_vector_gpu(A_gpu, row_start, row_end, part_col_start, part_col_end,
						     part_num_entries,
						     x_gpu.gpu_mem_ptr() + part_col_start,
						     b_gpu.gpu_mem_ptr() + row_start,
						     num_vecs, next_vec_offset, next_vec_offset, spmv_vorder);
	      
	      if(!success)
		goto done;
	   
	      index_type gpu_mem_offset_start = row_start;
	      index_type gpu_mem_offset_end   = row_end;
	      success = b_gpu.copy_to_host(b+row_start, gpu_mem_offset_start, gpu_mem_offset_end);
	      
	      if(!success)
		goto done;

	    }

	  //success = csr_tri_solve_part_cpu(A, row_start, row_end, uplo_type, &b_cp.front() + row_start, x + row_start,
	  //num_vecs, next_vec_offset, vorder);
	  
	  success = csr_tri_solve_part_cpu(A, nnzPerRowDiagBlock, row_start, row_end, uplo_type, b + row_start, x + row_start,
					   num_vecs, next_vec_offset, vorder);

	  if(!success)
	    goto done;
	  
	  index_type host_mem_offset_start = row_start;
	  index_type host_mem_offset_end   = row_end;
	  
	  success = x_gpu.copy_from_host(x, host_mem_offset_start, host_mem_offset_end);
	  
	  if(!success)
	    goto done;


	  if(!up)
	    ++i_b;
	}
      
      success = true;
    }

 done:

  return success;
}

// -----------------------------------------------------------------------------
template<typename index_type, typename value_type>
int csr_tri_solve_gpu_hybrid(const csr_matrix_data<index_type, value_type>* A,
			     index_type* nnzPerRowDiagBlock,
			     uplo_enum uplo_type,
			     value_type* b,
			     value_type* x,
			     const csr_matrix_data<index_type, value_type>* A_gpu,
			     value_type* b_gpu,
			     value_type* x_gpu,
			     index_type block_size,
			     index_type num_vecs,
			     unsigned int vorder)
{
  cuda_memory<value_type> b_gpu_wrap(b_gpu, A->num_entries * num_vecs, false);
  cuda_memory<value_type> x_gpu_wrap(x_gpu, A->num_entries * num_vecs, false);
  
  bool success = csr_tri_solve_gpu_hybrid(*A, nnzPerRowDiagBlock, uplo_type, b, x, *A_gpu, b_gpu_wrap, x_gpu_wrap,
					  block_size, num_vecs, vorder);
  
  return success ? 0 : 1;
}


#endif // CSR_TRI_SOLVE_GPU_H


